///////////////////////////////////////////////////////
// sq80dir - list contents of SQ80 disks             //
// written 1999 by Rainer Buchty (rainer@buchty.net) //
///////////////////////////////////////////////////////

#include <fcntl.h>
#include <linux/fd.h>
#include <stdio.h>
#include "sq80.h"

#define S_DISK 0
#define S_FILE 1

void usage(void){
	printf("sq80dir drive [-if=dumpfile] [-nX] [-dY]\n");
	printf("\tWhere \"drive\" can be:\n");
	printf("\ta:\tcomplete directory\n");
	printf("\tb:\tprogram banks\n");
	printf("\tp:\tsingle programs\n");
	printf("\ts:\tsequencer files\n");
	printf("\n\tOptional Parameters:\n");
	printf("\tdumpfile\t access dump file instead of disk\n");
	printf("\tX\t list file content (only for bank and song files)\n");
	printf("\tY\t select physical disk drive (0-3)\n");
	exit(0);
}

void error(const char *message){
	printf("%s\n",message);
	exit(-1);
}

void printchar(char text){
        printf("%c",text);
}

void undln(char text){
        printf("%s%c%s",SUL,text,EUL);
}

void sqprint(char *text,int len){
int i;
        for(i=0;i<len;i++){
                switch(text[i]){
			case 0x00:	printchar(0x2d);
					break;
                        case 0x21:      undln(0x30);
                                        break;
                        case 0x23:      undln(0x31);
                                        break;
                        case 0x25:      undln(0x32);
                                        break;
                        case 0x28:      undln(0x33);
                                        break;
                        case 0x29:      undln(0x34);
                                        break;
                        case 0x3a:      undln(0x35);
                                        break;
                        case 0x3b:      undln(0x36);
                                        break;
                        case 0x5b:      undln(0x37);
                                        break;
                        case 0x5c:      undln(0x38);
                                        break;
                        case 0x5d:      undln(0x39);
                                        break;   
                        default:        printchar(text[i]);
                                        break;
                }
        }
}

int translate_chs(int track,int head, int sector) {
int ptr=0;
        if (track>0) ptr=track*11264;                           /* complete track offset */
        if (head==1) ptr+=5632;                                 /* "half" track offset */
        if ((sector>0) && (sector!=5)) ptr+=sector*1024;        /* big sector offset */
        if (sector==5) ptr+=5120;                               /* small sector offset */
        return ptr+10;                                          /* header size offset */
}

void sqseek(int dnum, int c) {
char dummy[256];
	sprintf(dummy,"fdrawcmd seek %i %i 2>/dev/null",(dnum&3),c);
	system(dummy);
}

int main(int argc, char **argv){
char dummy[256],diskbuf[4096],inname[256];
FILE *fd;

int fnum=0, dnum=0, ftype=-1, i, pallow, source, ptr;
int c,h,s;

	argv++; argc--;
	
	if ((argc>3)||(argc<1)) usage();

	inname[0]=0;
	while(*argv!=NULL){
		if (strncmp(*argv,"-n",2)==0) fnum=strtoul((*argv)+2,0,0);
		else if (strncmp(*argv,"-d",2)==0) dnum=strtoul((*argv)+2,0,0);
		else if (strncmp(*argv,"-if=",4)==0) strcpy(inname,*argv+4);
		else if (strcmp(*argv,"a:")==0) ftype=ALL;
		else if (strcmp(*argv,"b:")==0) ftype=BNK;
		else if (strcmp(*argv,"p:")==0) ftype=PRG;
		else if (strcmp(*argv,"s:")==0) ftype=ASX;
		argv++;
	}

	if (inname[0]==0) source=S_DISK; else source=S_FILE;

	if ((dnum<0)||(dnum>3)) usage();

	if (ftype==-1) usage();

	if (fnum!=0) {
		if (ftype==BNK)
			if ((fnum<1)||(fnum>40)) error("Bank files range from 1 to 40.");
		if (ftype==ASX)
			if ((fnum<1)||(fnum>10)) error("Sequencer files range from 1 to 10.");
		if (ftype==PRG)
			fnum=0;
	}

	// get directory from disk //
	if (source==S_DISK) {
		// reset floppy //
		if ((i=open(floppy[dnum],O_NDELAY))<0) error("Cant access floppy drive.\n");
		if (ioctl(i,FDRESET,2)<0) {
			close(i);
			error("Cant reset floppy drive.");
		}
		close(i);

		// tune to SQ80 disk //
		sprintf(dummy,"fdrawcmd readid %i rate=2 2>/dev/null",dnum);
		system(dummy);

		sprintf(dummy,"fdrawcmd recalibrate %i 2>/dev/null",dnum);
		system(dummy);
	
		// read directory //
		for(i=0;i<4;i++) {
			c=dirpos[i].c; h=dirpos[i].h;
			if ((i%2)==0) sqseek(dnum,c);
			sprintf(dummy,"fdrawcmd read %i %i %i 5 2 0 0x1b 0xff length=512 rate=2 2>/dev/null",(dnum&3)|(h<<2),c,h);	
			fd=popen(dummy,"r");
				fread(dirbuf+i*512,1,512,fd);
			fclose(fd);
		}	

	// get directory from file //
	} else {
		if ((fd=fopen(inname,"r"))==NULL) error("Cant open file.\n");
		if (fread(dummy,1,10,fd)!=10) {
			fclose(fd);
			error("Short file.\n");
		}

		for(i=0;i<4;i++){
			ptr=translate_chs(dirpos[i].c,dirpos[i].h,dirpos[i].s);
			if (fseek(fd,ptr,SEEK_SET)<0) {
				fclose(fd);
				error("Short file.\n");
			}
			if (fread(dirbuf+i*512,1,512,fd)!=512) {
				fclose(fd);
				error("Short file.\n");		
			}
		}
		fclose(fd);
	}	
		
		
	// create directory structure //
	for(i=0;i<178;i++) {
		if (i<10) {				// sequencer files
			diskdir[i].type=dirbuf[i*13];
			diskdir[i].size=((dirbuf[i*13+11]&0xff)<<8)|(dirbuf[i*13+12]&0xff);	
			memcpy(diskdir[i].name,dirbuf+i*13+1,10);
			diskdir[i].name[10]=0;
		} else if ((i>9)&&(i<50)) {		// program banks
			diskdir[i].type=BNK;
			diskdir[i].size=4080;
			memcpy(diskdir[i].name,dirbuf+i*13+1,10);
			diskdir[i].name[10]=0;
		} else {				// single programs
			diskdir[i].type=PRG;
			diskdir[i].size=102;
			memcpy(diskdir[i].name,dirbuf+650+(i-50)*6,6);
			diskdir[i].name[6]=0;
		}
	}
	
	// list disk files //	
	if ( (fnum==0)||(ftype==PRG) ) {
		for(i=0;i<178;i++) {
			pallow=0;

			if ( (i<10) && ((ftype==ALL)||(ftype==ASX)) ) {
				pallow=1;
				if (i==0) printf("\n%sSequencer Files%s\n",SUL,EUL);
			}
			
			if ( ((i>9)&&(i<50)) && ((ftype==BNK)||(ftype==ALL)) ) {
				pallow=2;
				if (i==10) printf("\n%sProgram Banks%s\n",SUL,EUL);
			}

			if ( (i>49) && ((ftype==PRG)||(ftype==ALL)) ) {
				pallow=3;
				if (i==50) printf("\n%sSingle Programs%s\n",SUL,EUL);
			}					

			if (pallow==3) printf("%3i: ",i-49);
			else if (pallow==2) printf("%2i: ",i-9);
			else if (pallow==1) printf("%2i: ",i+1);
			
			if (pallow==3) sqprint(diskdir[i].name,6);
			else if (pallow!=0) sqprint(diskdir[i].name,10);

			if (pallow==1)
				printf(" .%s (%5i bytes)",ft_ext[diskdir[i].type],diskdir[i].size);

			if (pallow==1) 
				if ((i+1)%2==0) printf("\n"); else printf("\t");
			else if (pallow==2)
				if ((i-9)%5==0) printf("\n"); else printf("  ");
			else if (pallow==3)
				if ((i-49)%6==0) printf("\n"); else printf("  ");
		}
		if ((ftype==PRG)||(ftype==ALL)) printf("\n");

	// list file contents (SNG/BNK only) //
	} else {
		if (ftype==BNK)	{
			c=geo[fnum+9].st; h=geo[fnum+9].sh; s=geo[fnum+9].ss;

			// get bank from disk //
			if (source==S_DISK) {
#ifdef SLOW_FLOPPY
				for(i=0;i<=c;i++) sqseek(dnum,i);
#else
				sqseek(dnum,c);
#endif
				for(i=0;i<4;i++) {
					sprintf(dummy,"fdrawcmd read %i %i %i %i 3 0 0x1b 0xff length=1024 rate=2 2>/dev/null",(dnum&3)|(h<<2),c,h,s);	
					fd=popen(dummy,"r");
						fread(diskbuf+i*1024,1,1024,fd);
					fclose(fd);
					if (s<4) s++;
					else {
						s=0;
						c++;
						sqseek(dnum,c);
					}
				}

			// get bank from file //
			} else {
				for(i=0;i<4;i++) {
					ptr=translate_chs(c,h,s);
					if ((fd=fopen(inname,"r"))==NULL) error("Cant open file.\n");
					if (fseek(fd,ptr,SEEK_SET)<0) {	
						fclose(fd);
						error("Short file.\n");
					}
					if (fread(diskbuf+i*1024,1,1024,fd)!=1024) {
						fclose(fd);
						error("Short file.\n");
					}
					if (s<4) s++;
					else {
						s=0;
						c++;
					}
				}
					
			}

			printf("\n%sContents of bank %i (",SUL,fnum);
			sqprint(diskdir[fnum+9].name,10);
			printf(")%s\n",EUL);
			for(i=0;i<40;i++) {
				printf("%2i: ",i+1);
				memcpy(dummy,diskbuf+i*102,6);
				sqprint(dummy,6);
				if ((i+1)%5==0) printf("\n"); else printf("  ");
			}								

		// sequencer data //
		} else {
			if (diskdir[fnum-1].type==SNG) {
				c=geo[fnum-1].et; h=geo[fnum-1].eh;
	
				// get data from disk //
				if (source==S_DISK) {
#ifdef SLOW_FLOPPY
					for(i=0;i<=c;i++) sqseek(dnum,i);
#else	
					sqseek(dnum,c);
#endif
					sprintf(dummy,"fdrawcmd read %i %i %i 5 2 0 0x1b 0xff length=512 rate=2 2>/dev/null",(dnum&3)|(h<<2),c,h);	
					fd=popen(dummy,"r");
						fread(diskbuf,1,512,fd);
					fclose(fd);

				// get data from file //
				} else {
					ptr=translate_chs(c,h,5);
					if ((fd=fopen(inname,"r"))==NULL) error("Cant open file.\n");
					if (fseek(fd,ptr,SEEK_SET)<0) {	
						fclose(fd);
						error("Short file.\n");
					}
					if (fread(diskbuf,1,512,fd)!=512) {
						fclose(fd);
						error("Short file.\n");
					}
					fclose(fd);
				}				
			
				printf("\n%sContents of song %i (",SUL,fnum);
				sqprint(diskdir[fnum-1].name,10);
				printf(")%s\n",EUL);
				for(i=0;i<20;i++) {
					printf("%2i: ",i+1);
					memcpy(dummy,diskbuf+i*6+297,6);
					sqprint(dummy,6);	
					if ((i+1)%5==0) printf("\n"); else printf("  ");
				}
			} else {
				printf("\nFile %i (",fnum+1);
				sqprint(diskdir[fnum-1].name,10);
				printf(".%s) is no song.\n",ft_ext[diskdir[fnum-1].type]);
			}
		}
	}		

	sqseek(dnum,0);
	printf("\n");
	exit(0);
}
